Claw = {
    locked = false,
	scale = 1,
	frequency = 0.05,
	latency = 0.3,
	glow = true,
	charge = true,
	berserk = true,
	tfberserk = false,
	roar = true,
	motw = true,
}

if select(2, UnitClass("player")) ~= "DRUID" then
    DisableAddOn("Claw")
    return
end

SLASH_Claw1, SLASH_Claw2 = "/claw", "/clawesome"

local events = {}
local bestChoiceAbility = nil
local lastAbilityCalculation = 0
local targetHealthArray = {}
local playerHasShredGlyph = false
local ripExtendsLeft = 0
local clawPanel = CreateFrame("Frame", "clawPanel", UIParent)
clawPanel:SetPoint("CENTER", 0, -120)
clawPanel:SetFrameStrata("BACKGROUND")
clawPanel:Hide()

local Ability = {}
Ability.__index = Ability

function Ability.add(spellId, isDoT, playerCastOnly)
    local ability = {}
    setmetatable(ability, Ability)
    local spellInfo = {  }
	ability.spellId = spellId
    ability.isDoT = isDoT
    ability.playerCastOnly = playerCastOnly
    ability.name = GetSpellInfo(spellId)
    ability.icon = select(3, GetSpellInfo(spellId))
    return ability
end

function Ability:ready(seconds)
	return self:cooldown() <= (seconds or 0) + Claw.latency
end

function Ability:ticking()
    return UnitDebuff("target", self.name, nil, self.playerCastOnly and "PLAYER" or false) and true
end

function Ability:remains()
    return (self.isDoT and (select(7, UnitDebuff("target", self.name, nil, self.playerCastOnly and "PLAYER" or nil)) or GetTime()) or (select(7, UnitBuff("player", self.name, nil, self.playerCastOnly and "PLAYER" or nil)) or GetTime())) - GetTime()
end

function Ability:up()
    return UnitBuff("player", self.name, nil, self.playerCastOnly and "PLAYER" or false) and true
end

function Ability:inRange()
    return IsSpellInRange(self.name, "target") == 1
end

function Ability:cooldown()
    local cooldown = { GetSpellCooldown(self.name) }
    return cooldown[1] > 0 and cooldown[2] - (GetTime() - cooldown[1]) or 0
end

function Ability:stack()
    return self.isDoT and select(4, UnitDebuff("target", self.name, nil, self.playerCastOnly and "PLAYER" or nil)) or select(4, UnitBuff("player", self.name, nil, self.playerCastOnly and "PLAYER" or nil)) or 0
end

function Ability:known()
	return GetSpellBookItemInfo(self.name, "player") and true or false
end

-- Druid abilities
local Berserk = Ability.add(50334, false, true)
local FaerieFire = Ability.add(91565, true, false)
local FaerieFireFeral = Ability.add(16857, true, false)
local MarkOfTheWild = Ability.add(1126, false, false)
local OmenOfClarity = Ability.add(16870, false, true)
-- Cat abilities
local CatForm = Ability.add(768, false, true)
local FeralChargeCat = Ability.add(49376, false, true)
local FerociousBite = Ability.add(22568, false, true)
local MangleCat = Ability.add(33876, true, false)
local Rake = Ability.add(1822, true, true)
local Ravage = Ability.add(6785, false, true)
local Rip = Ability.add(1079, true, true)
local SavageRoar = Ability.add(52610, false, true)
local Shred = Ability.add(5221, false, true)
local Stampede = Ability.add(81022, false, true)
local TigersFury = Ability.add(5217, false, true)
-- Bear abilities
local BearForm = Ability.add(5487, false, true)
local DemoralizingRoar = Ability.add(99, true, false)
local Enrage = Ability.add(5229, false, true)
local FeralChargeBear = Ability.add(16979, false, true)
local Lacerate = Ability.add(33745, true, true)
local MangleBear = Ability.add(33878, true, false)
local Maul = Ability.add(6807, false, true)
local Pulverize = Ability.add(80313, false, true)
local Thrash = Ability.add(77758, true, true)
-- Other class abilities
--   10% stat buff
local BlessingOfKings = Ability.add(79063, false, false)
--   10% damage reduction
local DemoralizingShout = Ability.add(1160, true, false)
local ScarletFever = Ability.add(81130, true, false)
local Vindication = Ability.add(26017, true, false)
--   12% armor reduction
local ExposeArmor = Ability.add(8647, true, false)
local SunderArmor = Ability.add(7386, true, false)
--   30% bleed damage
local Hemorrhage = Ability.add(16511, true, false)
local PetGore = Ability.add(35290, true, false)
local PetStampede = Ability.add(57386, true, false)
local PetTendonRip = Ability.add(50271, true, false)
local Trauma = Ability.add(46857, true, false)

local function ComboPoints()
    return GetComboPoints("player", "target")
end

local function Energy()
    return CatForm:up() and UnitPower("player") or 100
end

local function Rage()
    return BearForm:up() and UnitPower("player") or 0
end

local function TargetHealthPercentage()
    return UnitHealth("target") / UnitHealthMax("target") * 100
end

local function TargetTimeToDie()
    local targetTimeToDie = UnitHealth("target") / ((targetHealthArray[1] - targetHealthArray[math.floor(3 / Claw.frequency)]) / 3)
    return targetTimeToDie >= 0 and targetTimeToDie - Claw.latency or 1/0
end

local function TargetIsBoss()
	return UnitLevel("target") == -1 or (UnitLevel("target") == 87 and not UnitInRaid("player"))
end

local function CalculateBestChoiceAbilityCat()
    if not Shred:inRange() and (FaerieFireFeral:ready() and FaerieFire:stack() < 3) and not (ExposeArmor:ticking() or SunderArmor:ticking()) then
        return FaerieFireFeral
    elseif Claw.charge and FeralChargeCat:ready() and FeralChargeCat:inRange() then
        return FeralChargeCat
    elseif TargetTimeToDie() < 2 and ComboPoints() >= 1 and Energy() >= 25 then
        return FerociousBite
	elseif Claw.tfberserk and TigersFury:ready() and Berserk:ready() and Energy() <= 35 then
		return Berserk
    elseif TigersFury:ready() and Energy() <= 35 and not Berserk:up() then
        return TigersFury
    elseif (Claw.berserk or TargetIsBoss()) and Berserk:ready() and (TigersFury:up() or (TargetTimeToDie() < 25 and not TigersFury:ready(6))) then
        return Berserk
    elseif (FaerieFireFeral:ready() and FaerieFire:stack() < 3) and not (ExposeArmor:ticking() or SunderArmor:ticking()) then
        return FaerieFireFeral
    elseif MangleCat:remains() <= 2 + Claw.latency and not (Trauma:ticking() or Hemorrhage:ticking() or PetGore:ticking() or PetStampede:ticking() or PetTendonRip:ticking()) then
        return MangleCat
    elseif Stampede:up() and Stampede:remains() <= 1 + Claw.latency then
        return Ravage
    elseif ((ComboPoints() >= 1 and Rip:ticking() and Rip:remains() <= 1 + Claw.latency) or (ComboPoints() == 5 and Rip:ticking())) and TargetHealthPercentage() <= 25 then
        return FerociousBite
    elseif playerHasShredGlyph and ripExtendsLeft >= 1 and Rip:ticking() and Rip:remains() <= 4 + Claw.latency and TargetHealthPercentage() > 25 then
        return Shred
    elseif ComboPoints() == 5 and TargetTimeToDie() >= 6 + Claw.latency and Rip:remains() <= 2 + Claw.latency and (Berserk:up() or Rip:remains() <= TigersFury:cooldown()) then
        return Rip
    elseif ComboPoints() == 5 and Rip:remains() >= 5 + Claw.latency and SavageRoar:remains() >= 3 + Claw.latency and Berserk:up() then
        return FerociousBite
    elseif (TargetTimeToDie() >= 8.5 + Claw.latency and TigersFury:up() and Rake:remains() <= 9 + Claw.latency) or (TargetTimeToDie() >= Rake:remains() and Rake:remains() <= 3 + Claw.latency and (Berserk:up() or Energy() > 70 or TigersFury:cooldown() + 0.8 >= Rake:remains())) then
        return Rake
    elseif OmenOfClarity:up() then
        return Shred
    elseif ComboPoints() >= 1 and SavageRoar:remains() <= 1 + Claw.latency then
        return SavageRoar
    elseif ComboPoints() == 5 and (TargetTimeToDie() <= 4 + Claw.latency or (Rip:remains() >= 11 + Claw.latency and SavageRoar:remains() >= 8 + Claw.latency)) or TargetTimeToDie() <= 1 + Claw.latency then
        return FerociousBite
    elseif Stampede:up() and not OmenOfClarity:up() and TigersFury:up() then
        return Ravage
    elseif TigersFury:up() or Berserk:up() or (ComboPoints() < 5 and Rip:remains() <= 3 + Claw.latency) or (ComboPoints() == 0 and SavageRoar:remains() <= 2 + Claw.latency) or TigersFury:ready(3) or TargetTimeToDie() <= 8.5 + Claw.latency or Energy() >= 78 then
        return Shred
    else
        return nil
    end
end

local function CalculateBestChoiceAbilityBear()
    if (FaerieFireFeral:ready() and FaerieFire:stack() < 3) and not (ExposeArmor:ticking() or SunderArmor:ticking()) then
        return FaerieFireFeral
	elseif Enrage:ready() and Rage() <= 80 then
		return Enrage
    elseif Claw.charge and FeralChargeBear:ready() and FeralChargeBear:inRange() then
        return FeralChargeBear
	elseif Maul:ready() and Rage() >= 70 then
		return Maul
	elseif MangleBear:ready(0.4) then
		return MangleBear
	elseif not Lacerate:ticking() or (Lacerate:ticking() and Lacerate:remains() <= 2 + Claw.latency) then
		return Lacerate
	elseif Claw.roar and not (DemoralizingRoar:ticking() or DemoralizingShout:ticking() or Vindication:ticking() or ScarletFever:ticking()) then
		return DemoralizingRoar
	elseif Thrash:ready(0.4) then
		return Thrash
	elseif Lacerate:stack() < 3 then
		return Lacerate
	elseif Pulverize:known() and Lacerate:stack() == 3 and (not Pulverize:up() or (Pulverize:up() and Pulverize:remains() <= 2 + Claw.latency)) then
		return Pulverize
	elseif FaerieFireFeral:ready() then
		return FaerieFireFeral
	else
        return Lacerate
    end
end

local function CalculateBestChoiceAbility()
    if Claw.motw and not (MarkOfTheWild:up() or BlessingOfKings:up()) then
        return MarkOfTheWild
    elseif CatForm:up() then
        return CalculateBestChoiceAbilityCat()
	elseif BearForm:up() then
		return CalculateBestChoiceAbilityBear()
	elseif Pulverize:known() then
		return BearForm
	else
		return CatForm
	end
end

local function UpdateAbilityButtonGlow(actionBar)
	local actionButtons = { actionBar:GetChildren() }
	local actioninfo = nil
	for i = 1, 12 do
		actionInfo = { GetActionInfo(actionButtons[i].action) }
		if bestChoiceAbility and Claw.glow and ((actionInfo[1] == "spell" and actionInfo[2] == bestChoiceAbility.spellId) or (actionInfo[1] == "macro" and select(2, GetMacroInfo(actionInfo[2])) == bestChoiceAbility.icon)) then
			ActionButton_ShowOverlayGlow(actionButtons[i])
		else
			ActionButton_HideOverlayGlow(actionButtons[i])
		end
	end
end

local function UpdateAbilityBars()
	UpdateAbilityButtonGlow(MultiBarRight)
	UpdateAbilityButtonGlow(MultiBarLeft)
	UpdateAbilityButtonGlow(MultiBarBottomLeft)
	UpdateAbilityButtonGlow(MultiBarBottomRight)
	UpdateAbilityButtonGlow(BonusActionBarFrame)
end

function events:PLAYER_TARGET_CHANGED(source)
    if UnitCanAttack("player", "target") and not UnitIsDead("target") then
		clawPanel:Show()
        for i = 1, math.floor(3 / Claw.frequency) do
            targetHealthArray[i] = UnitHealth("target")
        end
    else
        targetHealthArray = {}
        clawPanel:Hide()
		bestChoiceAbility = nil
		if Claw.glow then
			UpdateAbilityBars()
		end
    end
end

function events:UNIT_FACTION(unit)
	if unit == "target" then
		events:PLAYER_TARGET_CHANGED()
	end
end

function events:ACTIVE_TALENT_GROUP_CHANGED(index)
	playerHasShredGlyph = false
	for i = 1, GetNumGlyphSockets() do
		if select(4, GetGlyphSocketInfo(i)) == 54815 then
			playerHasShredGlyph = true
			break
		end
	end
end

function events:ADDON_LOADED(name)
	if name == "Claw" then
		clawPanel:SetWidth(64 * Claw.scale)
		clawPanel:SetHeight(64 * Claw.scale)
		clawPanel:RegisterForDrag("LeftButton")
		clawPanel:SetScript("OnDragStart", clawPanel.StartMoving)
		clawPanel:SetScript("OnDragStop", clawPanel.StopMovingOrSizing)
		clawPanel:SetMovable(true)
		clawPanel:EnableMouse(not Claw.locked)
		events:ACTIVE_TALENT_GROUP_CHANGED(0)
	end
end

function events:COMBAT_LOG_EVENT_UNFILTERED(self, eventType, hideCaster, srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, dstRaidFlags, spellId, spellName, spellSchool, auraType)
    if srcGUID == UnitGUID("player") and dstGUID == UnitGUID("target") then
        if spellId == 1079 and (eventType == "SPELL_AURA_APPLIED" or eventType == "SPELL_AURA_REFRESH") then
            ripExtendsLeft = 3
        elseif spellId == 5221 and eventType == "SPELL_DAMAGE" then
            ripExtendsLeft = ripExtendsLeft - 1
        end
    end
end

clawPanel:SetScript("OnUpdate", function(self, elapsed)
    lastAbilityCalculation = lastAbilityCalculation + elapsed
    if lastAbilityCalculation >= Claw.frequency then
        table.insert(targetHealthArray, UnitHealth("target"))
        table.remove(targetHealthArray, 1)
		local oldBestChoiceAbility = bestChoiceAbility
        bestChoiceAbility = CalculateBestChoiceAbility()
		if oldBestChoiceAbility ~= bestChoiceAbility then
			clawPanel:SetBackdrop({ bgFile = bestChoiceAbility and bestChoiceAbility.icon or nil; })
			if Claw.glow then
				UpdateAbilityBars()
			end
		end
        lastAbilityCalculation = 0
    end
end)

clawPanel:SetScript("OnEvent", function(self, event, ...) events[event](self, ...) end)
for k, v in pairs(events) do
    clawPanel:RegisterEvent(k)
end

function SlashCmdList.Claw(msg, editbox)
    msg = { strsplit(" ", strlower(msg)) }
    if msg[1] == "locked" then
        if msg[2] then Claw.locked = msg[2] == "on" end
		clawPanel:EnableMouse(not Claw.locked)
        print("Claw - Locked: " .. (Claw.locked and "|cFF00C000On" or "|cFFC00000Off"))
    elseif msg[1] == "charge" then
        if msg[2] then Claw.charge = msg[2] == "on" end
        print("Claw - Feral Charge: " .. (Claw.charge and "|cFF00C000On" or "|cFFC00000Off"))
    elseif msg[1] == "berserk" then
        if msg[2] then Claw.berserk = msg[2] == "on" end
        print("Claw - Berserk on non-boss mobs: " .. (Claw.berserk and "|cFF00C000On" or "|cFFC00000Off"))
	elseif msg[1] == "tfberserk" then
        if msg[2] then Claw.tfberserk = msg[2] == "on" end
        print("Claw - Tiger's Fury is macro'd to Berserk: " .. (Claw.tfberserk and "|cFF00C000On" or "|cFFC00000Off"))
    elseif msg[1] == "roar" then
        if msg[2] then Claw.roar = msg[2] == "on" end
        print("Claw - Demoralizing Roar (Bear Form): " .. (Claw.roar and "|cFF00C000On" or "|cFFC00000Off"))
    elseif msg[1] == "motw" then
        if msg[2] then Claw.motw = msg[2] == "on" end
        print("Claw - Mark of the Wild: " .. (Claw.motw and "|cFF00C000On" or "|cFFC00000Off"))
    elseif msg[1] == "glow" then
        if msg[2] then Claw.glow = msg[2] == "on" end
		UpdateAbilityBars()
        print("Claw - Glowing Ability Buttons: " .. (Claw.glow and "|cFF00C000On" or "|cFFC00000Off"))
	elseif msg[1] == "scale" then
		Claw.scale = tonumber(msg[2]) or 1
		clawPanel:SetWidth(64 * Claw.scale)
		clawPanel:SetHeight(64 * Claw.scale)
		print("Claw - Scale: |cFFFFD000" .. Claw.scale .. "|r times 64px")
	elseif msg[1] == "frequency" then
		Claw.frequency = tonumber(msg[2]) or 0.05
		print("Claw - Calculation frequency: |cFFFFD000" .. Claw.frequency .. "|r seconds")
	elseif msg[1] == "latency" then
		Claw.latency = tonumber(msg[2]) or 0.2
		print("Claw - Player reaction time: |cFFFFD000" .. Claw.latency .. "|r seconds")
    elseif msg[1] == "reset" then
        clawPanel:SetPoint("CENTER", 0, -120)
    elseif msg[1] == "" then
        print("Claw (version: |cFFFFD000" .. GetAddOnMetadata("Claw", "Version") .. "|r) - Commands:")
        print("  /claw locked |cFF00C000on|r/|cFFC00000off|r - lock the Claw UI so that it can't be moved")
		print("  /claw charge |cFF00C000on|r/|cFFC00000off|r - enable usage of Feral Charge")
		print("  /claw tfberserk |cFF00C000on|r/|cFFC00000off|r - Tiger's Fury is macro'd to Berserk (Cat Form)")
		print("  /claw berserk |cFF00C000on|r/|cFFC00000off|r - enable usage of Berserk on non-boss mobs (Cat Form)")
		print("  /claw roar |cFF00C000on|r/|cFFC00000off|r - enable usage of Demoralizing Roar (Bear Form)")
		print("  /claw motw |cFF00C000on|r/|cFFC00000off|r - enable usage of Mark of the Wild")
		print("  /claw glow |cFF00C000on|r/|cFFC00000off|r - enable glowing ability buttons on action bars")
		print("  /claw scale |cFFFFD000[number]|r - set the scale of the Claw UI (default is 1.0 times 64px)")
		print("  /claw frequency |cFFFFD000[number]|r - set the calculation frequency (default is 0.05 seconds)")
		print("  /claw latency |cFFFFD000[number]|r - set the latency of the player's reaction time (default is 0.2 seconds)")
        print("  /claw reset - reset the location of the Claw UI to default")
    end
end
